"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createConnectionPool = void 0;
const Logger_1 = require("../Logger");
const utilities_1 = require("@slonik/utilities");
const promises_1 = require("node:timers/promises");
const serialize_error_1 = require("serialize-error");
const logger = Logger_1.Logger.child({
    namespace: 'createConnectionPool',
});
const createConnectionPool = ({ driver, events, maximumPoolSize, minimumPoolSize, }) => {
    // See test "waits for all connections to be established before attempting to terminate the pool"
    // for explanation of why `pendingConnections` is needed.
    const pendingConnections = [];
    const connections = [];
    const waitingClients = [];
    const id = (0, utilities_1.generateUid)();
    let isEnding = false;
    let isEnded = false;
    let poolEndPromise = null;
    const endPool = async () => {
        try {
            await Promise.all(pendingConnections);
        }
        catch (error) {
            logger.error({
                error: (0, serialize_error_1.serializeError)(error),
            }, 'error in pool termination sequence while waiting for pending connections to be established');
        }
        try {
            await Promise.all(waitingClients.map((waitingClient) => waitingClient.deferred.promise));
        }
        catch (error) {
            logger.error({
                error: (0, serialize_error_1.serializeError)(error),
            }, 'error in pool termination sequence while waiting for waiting clients to be resolved');
        }
        // This is needed to ensure that all pending connections were assigned a waiting client.
        // e.g. "waits for all connections to be established before attempting to terminate the pool" test
        await (0, promises_1.setTimeout)(0);
        // Make a copy of `connections` array as items are removed from it during the map iteration.
        // If `connections` array is used directly, the loop will skip some items.
        await Promise.all([...connections].map((connection) => connection.destroy()));
    };
    const acquire = async () => {
        if (isEnding) {
            throw new Error('Connection pool is being terminated.');
        }
        if (isEnded) {
            throw new Error('Connection pool has ended.');
        }
        const addConnection = async () => {
            const pendingConnection = driver
                .createClient()
                // eslint-disable-next-line promise/prefer-await-to-then
                .then((resolvedConnection) => {
                return {
                    ...resolvedConnection,
                    events,
                };
            });
            pendingConnections.push(pendingConnection);
            const connection = await pendingConnection.catch((error) => {
                const index = pendingConnections.indexOf(pendingConnection);
                if (index === -1) {
                    logger.error('Unable to find pendingConnection in `pendingConnections` array to remove.');
                }
                else {
                    pendingConnections.splice(index, 1);
                }
                throw error;
            });
            const onRelease = () => {
                if (connection.state() !== 'IDLE') {
                    return;
                }
                const waitingClient = waitingClients.shift();
                if (!waitingClient) {
                    return;
                }
                connection.acquire();
                waitingClient.deferred.resolve(connection);
            };
            connection.on('release', onRelease);
            const onDestroy = () => {
                connection.removeListener('release', onRelease);
                connection.removeListener('destroy', onDestroy);
                const indexOfConnection = connections.indexOf(connection);
                if (indexOfConnection === -1) {
                    logger.error('Unable to find connection in `connections` array to remove.');
                }
                else {
                    connections.splice(indexOfConnection, 1);
                }
                const waitingClient = waitingClients.shift();
                if (waitingClient) {
                    // eslint-disable-next-line promise/prefer-await-to-then
                    acquire().then(waitingClient.deferred.resolve, waitingClient.deferred.reject);
                    return;
                }
                // In the case that there are no waiting clients and we're below the minimum pool size, add a new connection
                if (!isEnding && !isEnded && connections.length < minimumPoolSize) {
                    addConnection();
                }
            };
            connection.on('destroy', onDestroy);
            connections.push(connection);
            const indexOfPendingConnection = pendingConnections.indexOf(pendingConnection);
            if (indexOfPendingConnection === -1) {
                logger.error('Unable to find pendingConnection in `pendingConnections` array to remove.');
            }
            else {
                pendingConnections.splice(indexOfPendingConnection, 1);
            }
            return connection;
        };
        const idleConnection = connections.find((connection) => connection.state() === 'IDLE');
        if (idleConnection) {
            idleConnection.acquire();
            return idleConnection;
        }
        if (pendingConnections.length + connections.length < maximumPoolSize) {
            const newConnection = await addConnection();
            newConnection.acquire();
            return newConnection;
        }
        const deferred = (0, utilities_1.defer)();
        waitingClients.push({
            deferred,
        });
        const queuedAt = process.hrtime.bigint();
        logger.warn({
            connections: connections.length,
            maximumPoolSize,
            minimumPoolSize,
            pendingConnections: pendingConnections.length,
            waitingClients: waitingClients.length,
        }, `connection pool full; client has been queued`);
        // eslint-disable-next-line promise/prefer-await-to-then
        return deferred.promise.then((connection) => {
            logger.debug({
                connectionId: connection.id(),
                duration: Number(process.hrtime.bigint() - queuedAt) / 1e6,
            }, 'connection has been acquired from the queue');
            return connection;
        });
    };
    return {
        acquire,
        end: async () => {
            isEnding = true;
            if (poolEndPromise) {
                return poolEndPromise;
            }
            poolEndPromise = endPool();
            isEnded = true;
            return poolEndPromise;
        },
        id: () => {
            return id;
        },
        state: () => {
            const stateName = isEnded ? 'ENDED' : isEnding ? 'ENDING' : 'ACTIVE';
            const state = {
                acquiredConnections: 0,
                idleConnections: 0,
                pendingConnections: pendingConnections.length,
                pendingDestroyConnections: 0,
                pendingReleaseConnections: 0,
            };
            // TODO add pendingAcquireConnections
            // TODO add destroyedConnections
            for (const connection of connections) {
                if (connection.state() === 'ACQUIRED') {
                    state.acquiredConnections++;
                }
                if (connection.state() === 'IDLE') {
                    state.idleConnections++;
                }
                if (connection.state() === 'PENDING_RELEASE') {
                    state.pendingReleaseConnections++;
                }
                if (connection.state() === 'PENDING_DESTROY') {
                    state.pendingDestroyConnections++;
                }
                if (connection.state() === 'DESTROYED') {
                    state.pendingReleaseConnections++;
                }
            }
            return {
                ...state,
                state: stateName,
                waitingClients: waitingClients.length,
            };
        },
    };
};
exports.createConnectionPool = createConnectionPool;
//# sourceMappingURL=createConnectionPool.js.map